Package: GeoModels
Version: 2.2.8
Date: 2026-08-15
Depends: R (>= 4.1.0), fields, progressr, future.apply
Suggests: numDeriv, memuse, mapproj, scatterplot3d, sp
Authors@R: c(person("Moreno", "Bevilacqua", role = c("aut", "cre","cph"),
                     email = "moreno.bevilacqua89@gmail.com"),
             person("Víctor", "Morales-Oñate", role = "ctb"),
             person("Francisco", "Cuevas-Pacheco", role = "ctb"),
             person("Christian", "Caamaño-Carrillo", role = "ctb"))
Description: Functions for Gaussian and Non Gaussian (bivariate) spatial and spatio-temporal data analysis are provided for a) (fast) simulation of random fields,  b) inference  for random fields using standard likelihood and a likelihood approximation  method called  weighted composite likelihood based on pairs and b) prediction using (local) best linear unbiased prediction. Weighted composite likelihood can be very efficient for estimating massive datasets. Both regression and spatial (temporal) dependence analysis can be jointly performed. Flexible covariance models for spatial and spatial-temporal data on Euclidean domains and spheres are provided. There are also many useful functions for plotting and performing diagnostic analysis. Different non Gaussian random fields can be considered in the analysis. Among them, random fields with marginal distributions such as Skew-Gaussian, Student-t, Tukey-h, Sin-Arcsin, Two-piece, Weibull, Gamma, Log-Gaussian, Binomial, Negative Binomial  and Poisson. See the URL for the papers associated with this package, as for instance, Bevilacqua and Gaetan (2015) <doi:10.1007/s11222-014-9460-6>, Bevilacqua et al. (2016) <doi:10.1007/s13253-016-0256-3>, Vallejos et al. (2020) <doi:10.1007/978-3-030-56681-4>, Bevilacqua et. al (2020) <doi:10.1002/env.2632>, Bevilacqua et. al (2021) <doi:10.1111/sjos.12447>, Bevilacqua et al. (2022) <doi:10.1016/j.jmva.2022.104949>, Morales-Navarrete et al. (2023) <doi:10.1080/01621459.2022.2140053>, and a large class of examples and tutorials.
Title: Procedures for Gaussian and Non Gaussian Geostatistical (Large)
        Data Analysis
License: GPL (>= 3)
Encoding: UTF-8
Imports: methods, spam, dotCall64, FastGP, pbivnorm, sn, nabor,
        hypergeo, VGAM, future, minqa
URL: https://vmoprojs.github.io/GeoModels-page/
BugReports: https://github.com/vmoprojs/GeoModels/issues
NeedsCompilation: yes
Author: Moreno Bevilacqua [aut, cre, cph],
  Víctor Morales-Oñate [ctb],
  Francisco Cuevas-Pacheco [ctb],
  Christian Caamaño-Carrillo [ctb]
Maintainer: Moreno Bevilacqua <moreno.bevilacqua89@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-21 05:46:27 UTC
Packaged: 2026-08-20 21:30:50 UTC; morenobevilacqua
