export(CheckBiv,CheckDistance,CheckSph,CheckST,	CkCorrModel,CkInput,CkLikelihood,CkModel,	CkType,CompIndLik2,CompLik,CompLik2,	CorrelationPar,	CorrParam, corrsas, GeoAniso,
GeoKrigWeights,GeoKriglocWeights,GeoCorrFct_Cop,GeoDistances,	GeoCorrFct,GeoCovariogram,GeoCovDisplay,GeoCovmatrix,GeoCV,GeoFit,GeoFit2,GeoTestIsotropy,	GeoKrig,GeoKrigloc,GeoNA,GeoNeighborhood,GeoNeighIndex,GeoOutlier,GeoPit,	GeoQQ,GeoResiduals,GeoScatterplot,GeoSim,GeoSimapprox,GeoSimcond,GeoSimCopula,GeoScores,GeoTests,GeoVarestbootstrap,GeoVariogram,GeoVariogramDir,Lik,MatDecomp,NuisParam,NuisParam2,MargParam,GeoDoScores,StartParam,WlsStart,MatSqrt,MatInv,MatLogDet,GeoWLS,sp2Geo,GeoNeighbSelect,GeoTestsupp_space)
useDynLib(GeoModels, .registration = TRUE)
S3method(print,GeoFit) 
S3method(print,GeoCovmatrix) 
S3method(summary, GeoFit)
S3method(print,GeoWLS)
S3method(plot,GeoVariogram)
S3method(plot,GeoVariogramDir)
S3method(plot,GeoCorrFct)
importFrom("grDevices", dev.cur,gray.colors)
importFrom("graphics", abline, lines, par, persp,plot,points,image,plot.default,box,
axis,plot.new,plot.window,title)
importFrom("graphics", "hist","legend")
importFrom("stats", setNames, median, nlminb, dist, fft, quantile, qqnorm, qbinom,qpois,qnbinom,  pweibull, pgamma, plnorm, nlm, optim, optimize, integrate,approxfun,pchisq,pnorm,rt,qnorm,dnorm,dpois,rbeta,dbeta,dnbinom,dbinom,dweibull,dgamma,dlnorm,pbinom, plogis, pnbinom, ppois,rbinom,sd,
pbeta,dt,dlogis,qlnorm,qgamma,qweibull, qbeta,qlogis,lm, rnorm, runif, rgamma,rf,pt,qt,qunif,uniroot, var,cor)
importFrom("methods","is", "new", "slot<-","as")
importFrom("utils", "setTxtProgressBar", "txtProgressBar","capture.output")
importFrom("utils", "head","object.size")
importFrom("withr", with_seed)
importFrom("sp", "coordinates","is.projected")
importFrom("minqa","bobyqa")
importFrom("plotrix", "draw.circle")
importFrom("FastGP","rcppeigen_get_chol","rcppeigen_invert_matrix")
importFrom("VGAM","qzinegbin","qzipois")
import("fields", "mapproj","shape")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(future.apply, future_lapply)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future,sequential)
importFrom(doFuture,"%dofuture%")
importFrom(progressr,handlers,progressor)






