Test data for GeographicLib
===========================

*NOTE:* It's not necessary to download these files to use
GeographicLib.

`GeodTest.dat.gz` and `GeodTest-short.dat.gz` provide test data for
geodesics.  See

> https://geographiclib.sourceforge.io/C++/doc/geodesic.html#testgeod

`TMCoords.dat.gz` provides test data for the transverse Mercator
projection.  See

> https://geographiclib.sourceforge.io/C++/doc/transversemercator.html#testmerc

`GeoidHeights.dat.gz` provides test data for geoid heights.  See

> https://geographiclib.sourceforge.io/C++/doc/geoid.html#testgeoid
