Geoid models for use with GeographicLib
=======================================

These files give the height of geoid above the ellipsoid on a
rectangular grid.  It's necessary to download and install at least one
of these files if you want to use the `Geoid` class in GeographicLib
or the `geoid_height` function in the [geographiclib-octave](
https://github.com/geographiclib/geographiclib-octave#readme) package.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> egm96-5

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/geoid.html#geoidinst
