Gravity models for use with GeographicLib
=========================================

These files give the data for various gravity models.  It's
necessary to download and install at least one of these files if you
want to use the `GravityModel` class in GeographicLib.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> egm96

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/gravity.html#gravityinst
