# Source code for C++ version of GeographicLib

Programmers should download the latest `.tar.gz` or `.zip` file.  The
`.exe` files provide Windows installers for the utilities (and the
library compiled with Visual Studio 14 2015).

The algorithms are documented in

* C. F. F. Karney,
  [Algorithms for geodesics](https://doi.org/10.1007/s00190-012-0578-z),
  J. Geodesy 87, 43-55 (2013);
  [Addenda](https://geographiclib.sourceforge.io/misc/geod-addenda.html).

Other links:

* Installation instructions:
  https://geographiclib.sourceforge.io/C++/doc/install.html
* Library documentation: https://geographiclib.sourceforge.io/C++/doc
* Git repository: https://github.com/geographiclib/geographiclib
* Git mirror site: https://sourceforge.net/p/geographiclib/code
* Source distribution (2.0 and later):
  https://sourceforge.net/projects/geographiclib/files/distrib-C++
* GeographicLib: https://geographiclib.sourceforge.io
* GeographicLib in various languages:
  https://geographiclib.sourceforge.io/doc/library.html#languages
* Old versions of Geographiclib:
  https://sourceforge.net/projects/geographiclib/files/distrib
* Author: Charles Karney, <karney@alum.mit.edu>
