Magnetic models for use with GeographicLib
==========================================

These files give the data for various geomagnetic models.  It's
necessary to download and install at least one of these files if you
want to use the `MagneticModel` class in GeographicLib.

The data is provided either as a compressed tar (extension `.tar.bz2`), a
zip file, or a windows installer (extension `.exe`).  If you're unsure
which one to install, start with

> wmm2025

For more information, see

> https://geographiclib.sourceforge.io/C++/doc/magnetic.html#magneticinst
